/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: perf.c,v 1.25 2010/01/09 07:30:27 zlb Exp $
 *
 * Functions providing performance information */

#include "phg.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>	/* for strerror_r() */
#include <errno.h>

/* for open/read/close */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <signal.h>
#include <sys/time.h>
#ifdef USETIMES
# include <sys/times.h>
# include <unistd.h>
#else
# include <sys/resource.h>
#endif

/*---------------------- Memory usage statistics ------------------------*/

/* Real time */
#define TimerSignal	SIGALRM
#define TimerType	ITIMER_REAL

/* Virtual time */
/*#define TimerSignal	SIGVTALRM
#define TimerType	ITIMER_VIRTUAL*/

/* Profiling time */
/*#define TimerSignal	SIGPROF
#define TimerType	ITIMER_PROF*/

static size_t	mem_peak = 0;
static size_t	mem_current = 0;
/* The sampling interval of the timer in ms, <= 0 disables the timer. */
static INT	sampling_interval = 200;
static char	statm[40];
static BOOLEAN	have_statm = FALSE;

static void
sample_mem_usage(int dummy)
{
    int fd;
    struct rusage RU;
    size_t mem;
    static BOOLEAN locked = FALSE;

    if (locked)
	return;
    locked = TRUE;

    /* TODO: configure checks for different systems */

    /* first, try /proc/pid/statm for Linux systems */
    if (have_statm && (fd = open(statm, O_RDONLY)) != -1) {
	int rsspages;
	static char buffer[256];
	/* see linux-2.6.15/Documentation/filesystems/proc.txt */
	rsspages = read(fd, buffer, sizeof(buffer) - 1);
	close(fd);
	buffer[rsspages] = '\0';
	sscanf(buffer, "%*d %d", &rsspages);
	mem = (size_t)rsspages * (size_t)getpagesize();
    }
    else {
	/* next, try getrusage() */
	getrusage(RUSAGE_SELF, &RU);
	mem = RU.ru_maxrss * (size_t)1024;
	/*mem = RU.ru_maxrss * getpagesize();*/
    }

    if (mem > mem_peak)
	mem_peak = mem;
    mem_current = mem;
    locked = FALSE;
}

void
phgMemoryUsageReset(void)
{
    mem_peak = 0;
}

size_t
phgMemoryUsage(GRID *g, size_t *peak)
{
    sample_mem_usage(0);

#if USE_MPI
    if (g != NULL && g->nprocs > 1 && !phgMasterSlave) {
	double a[2], b[2];
	a[0] = mem_current + 0.5;
	a[1] = mem_peak + 0.5;
	MPI_Allreduce(a, b, 2, MPI_DOUBLE, MPI_MAX, g->comm);
	if (peak != NULL)
	    *peak = (size_t)b[1];
	return (size_t)b[0];
    }
#else	/* USE_MPI */
    Unused(g);
#endif	/* USE_MPI */

    if (peak != NULL)
	*peak = mem_peak;

    return mem_current;
}

size_t
phgMemoryPeakReset(size_t value)
{
    size_t old_value = mem_peak;
    mem_peak = value;

    return old_value;
}

size_t
phgMemoryPeakRestore(size_t value)
{
    size_t old_value = mem_peak;

    if (mem_peak < value)
	mem_peak = value;

    return old_value;
}

static BOOLEAN timer_initialized = FALSE;
static struct itimerval new, old;
static struct sigaction sa, old_sa;

void
phgTimerInit(BOOLEAN init)
{
    int fd;

    if (init) {
	phgOptionsRegisterInt("mem_sampling_rate", "Memory usage sampling "
		"rate (milliseconds, 0 disables memory sampling)",
		&sampling_interval);
	return;
    }

    if (timer_initialized)
	return;
    timer_initialized = TRUE;

    sprintf(statm, "/proc/%d/statm", (int)getpid());
    if ((fd = open(statm, O_RDONLY)) != -1) {
	have_statm = TRUE;
	close(fd);
    }

    if (sampling_interval <= 0)
	return;

    /* setup timer to sample memory usage */
    sa.sa_handler = sample_mem_usage;
    sigemptyset (&sa.sa_mask);
    /*sigfillset (&sa.sa_mask);*/
    sa.sa_flags = SA_RESTART;
    if (sigaction(TimerSignal, &sa, &old_sa))
	perror("sigaction 0");
    new.it_value.tv_sec = sampling_interval / 1000;
    new.it_value.tv_usec = (sampling_interval % 1000) * 1000;
    new.it_interval = new.it_value;
    if (setitimer(TimerType, &new, &old))
	perror("setitimer 0");

    return;
}

void
phgTimerReset(void)
{
    if (!timer_initialized)
	return;
    timer_initialized = FALSE;
 
    /* restore timer */
    if (setitimer(ITIMER_REAL, &old, NULL))
	perror("setitimer 1");
    if (sigaction(SIGALRM, &old_sa, NULL))
	perror("sigaction 1");

    return;
}

/*---------------------- Floating point operations ------------------------*/

static BOOLEAN no_perfctr = FALSE;

#if USE_PAPI

#include <papiStdEventDefs.h>
#include <papi.h>

static int (*func)(float *rtime, float *ptime, long_long *count, float *mf)
	= NULL;

static long_long count = 0;
static float rtime, ptime, mf;

static double time0 = -1., elapsed_time, last_time;
static long_long last_count = 0;

static void
set_time_count(void)
{
    double tarray[3];
    long_long tmp;

    phgGetTime(tarray);
    if (time0 < 0)
	time0 = tarray[2];
    elapsed_time = tarray[2] - last_time;
    tmp = count - last_count;
    last_count = count;
    last_time = tarray[2];
    count = tmp;
}

static void
papi_error(const char *funcname, int code)
{
    char errmsg[PAPI_MAX_STR_LEN];

    PAPI_perror(code, errmsg, sizeof(errmsg));
#if 0
    fprintf(stderr, "Initialization of PAPI_flops failed: %s", errmsg);
    if (code == PAPI_ESYS)
	perror(" ");
    else
	fprintf(stderr, "\n");
#else
    /*if (code == PAPI_ESYS) {
	char errmsg1[PAPI_MAX_STR_LEN];
	phgInfo(0, "Initialization of %s failed: %s %s\n", funcname, errmsg,
		strerror_r(errno, errmsg1, sizeof(errmsg1)));
    }
    else*/ {
	phgInfo(0, "Initialization of %s failed: %s\n", funcname, errmsg);
    }
#endif
}

void
phgPerfInit(void)
{
    int code;
    static int flag = -1;

    if (flag < 0) {
	/* register options */
	phgOptionsRegisterNoArg("no_perfctr", "Disable performance counters",
				&no_perfctr);
	flag = 0;
	return;
    }

    if (flag > 0 || no_perfctr)
	return;

    flag = 1;

#if 0
    PAPI_set_domain(PAPI_DOM_ALL);
    PAPI_set_granularity(/*PAPI_GRN_SYS*/PAPI_GRN_SYS_CPU);
#endif
    func = PAPI_flops;
    code = func(NULL, NULL, NULL, NULL);
    if (code != PAPI_OK) {
	papi_error("PAPI_flops", code);
	/*fprintf(stderr, "Trying PAPI_flips ...\n");*/
	func = PAPI_flips;
	code = func(NULL, NULL, NULL, NULL);
	if (code != PAPI_OK) {
	    papi_error("PAPI_flips", code);
	    func = NULL;
	}
    }

    if (func != NULL)
	set_time_count();
}

void
phgPerfFinalize(void)
{
    return;
}

double
phgPerfGetMflops(GRID *g, double *aggregate, double *since_last_call)
/* 'aggregate' = Mflops since the start of the program,
 * 'since_last_call' = Mflops since last call to this function.
 *
 * returns '*since_last_call' */
{
    double mflops[2];

    if (func == NULL) {
	if (aggregate != NULL)
	    *aggregate = 0.;
	if (since_last_call != NULL)
	    *since_last_call = 0.;
	return 0.;
    }

    func(&rtime, &ptime, &count, &mf);
    set_time_count();
    mflops[0] = (elapsed_time == 0.0 ? 0. : count / elapsed_time * 1e-6);
    elapsed_time = last_time - time0;
    count = last_count;
    mflops[1] = (elapsed_time == 0.0 ? 0. : count / elapsed_time * 1e-6);

#if USE_MPI
    if (g != NULL && g->nprocs > 1 && !phgMasterSlave) {
	double a[2];
	MPI_Allreduce(mflops, a, 2, MPI_DOUBLE, MPI_SUM, g->comm);
	mflops[0] = a[0];
	mflops[1] = a[1];
    }
#endif	/* USE_MPI */

    if (aggregate != NULL)
	*aggregate = mflops[0];

    if (since_last_call != NULL)
	*since_last_call = mflops[1];

    return mflops[0];
}

#else	/* USE_PAPI */

void
phgPerfInit(void)
{
    static int flag = -1;

    if (flag < 0) {
	/* register options */
	phgOptionsRegisterNoArg("no_perfctr", "disable performance counters",
				&no_perfctr);
	flag = 0;
	return;
    }
}

void
phgPerfFinalize(void)
{
    return;
}

double
phgPerfGetMflops(GRID *g, double *aggregate, double *since_last_call)
{
    if (aggregate != NULL)
	*aggregate = 0.;
    if (since_last_call != NULL)
	*since_last_call = 0.;

    return 0.;
}

#endif	/* USE_PAPI */
